/*
 *
 * Copyright (C) 2004 Mekensleep
 *
 *	Mekensleep
 *	24 rue vieille du temple
 *	75004 Paris
 *       licensing@mekensleep.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Authors:
 *  Loic Dachary <loic@gnu.org>
 *  Cedric Pinson <cpinson@freesheep.org>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <cassert>
#include <iostream>

#include <osg/Node>
#include <osg/Group>
#include <osg/MatrixTransform>
#include <osg/Geode>
#include <osg/NodeVisitor>
#include <osg/Geometry>
#include <osg/Material>
#include <osg/Texture2D>
#include <osg/StateSet>
#include <osg/Light>
#include <osg/LightSource>
#include <osgDB/ReadFile>

#include <exg/exg_visitor.h>
#include <exg/exg_mesh.h>
#include <exg/exg_python_scene.h>

#include <exg_osg.h>
#include <vector>
#include <string>

namespace exg
{
  class PyScene {
  public:
    PyScene() {
      mModule = PyImport_ImportModule("exg.scene");
      if(!mModule)
	abort();
    }

    PyObject* GetClass(const std::string& classname) {
      if(PyObject_HasAttrString(mModule, (char*)classname.c_str())) {
	return PyObject_GetAttrString(mModule, (char*)classname.c_str());
      } else {
	abort();
	return 0;
      }
    }


    PyObject* NodeNew(const std::string& classname, const std::string& instancename) {
      PyObject* pyclass = GetClass(classname);
      PyObject* instance = PyObject_CallFunction(pyclass, "s", (char*)instancename.c_str());
      if(!instance)
	abort();
      return instance;
    }

    bool IsInstance(PyObject* pyobject, const std::string& classname) {
      PyObject* pyclass = GetClass(classname);
      return PyObject_IsInstance(pyobject, pyclass) ? true : false;
    }

    osg::Vec3 Vector2osgVec3(PyObject* pyobject, const std::string& attribute) {
      osg::Vec3 vector;
      if(!PyObject_HasAttrString(pyobject, (char*)attribute.c_str())) {
	std::cerr << "Vector2osgVec3: attribute " << attribute << " not found" << std::endl;
	return vector;
      }

      PyObject* pyvector = PyObject_GetAttrString(pyobject, (char*)attribute.c_str());
      if(!PySequence_Check(pyvector)) {
	std::cerr << "Vector2osgVec3: expected sequence" << std::endl;
	return vector;
      }

      if(PySequence_Length(pyvector) != 3) {
	std::cerr << "Vector2osgVec3: expected sequence of size 3" << std::endl;
	return vector;
      }

      return osg::Vec3(PyFloat_AsDouble(PyNumber_Float(PySequence_GetItem(pyvector, 0))),
		       PyFloat_AsDouble(PyNumber_Float(PySequence_GetItem(pyvector, 1))),
		       PyFloat_AsDouble(PyNumber_Float(PySequence_GetItem(pyvector, 2))));
    }

  private:
    PyObject* mModule;
  };


  class EXGConverter : public Visitor
  {

   public:

    EXGConverter() : Visitor(ONCE), mGeode(0), mGeometry(0) {}

    void SetArgs(const std::vector<std::string>& args) {
      mArguments=args;
//       for (int i=0;i<mArguments.size();i++)
// 	std::cout << mArguments[i] << std::endl;

      // search where source data come from
      for (int i=0;i<mArguments.size();i++)
	if (mArguments[i].find(".escn")!=std::string::npos) {
	  std::string::size_type end=mArguments[i].rfind("/");
	  if (end!=std::string::npos) {
	    mSourceData=mArguments[i].substr(0,end+1);
	    //	    std::cout << "Data directory " << mSourceData << std::endl;
	    break;
	  }
	}
    }

    void Apply(Object& node) {
      std::cerr << "exg::EXGConverter " << node.ClassName() << " not implemented" << std::endl;
    }

    void Apply(Material& node) {
      if (!node.GetNbPolygons())
	return;
      mGeometry = new osg::Geometry;
      osg::ref_ptr<osg::StateSet> state = new osg::StateSet;
      osg::ref_ptr<osg::Material> material = new osg::Material;
      state->setAttribute(material.get());

      bool state_changed = false;
      VectorFloat* color;
      if (node.GetProperty("ambient")) {
	color = node.GetProperty("ambient")->AsVectorFloat();
	if(color) {
	  material->setAmbient(osg::Material::FRONT_AND_BACK, osg::Vec4((*color)[0], (*color)[1], (*color)[2], (*color)[3]));
	  state_changed = true;
	}
      }
      if (node.GetProperty("diffuse")) {
	color = node.GetProperty("diffuse")->AsVectorFloat();
	if(color) {
	  material->setDiffuse(osg::Material::FRONT_AND_BACK, osg::Vec4((*color)[0], (*color)[1], (*color)[2], (*color)[3]));
	  state_changed = true;
	}
      }

      if (node.GetProperty("specular")) {
	color = node.GetProperty("specular")->AsVectorFloat();
	if(color) {
	  material->setSpecular(osg::Material::FRONT_AND_BACK, osg::Vec4((*color)[0], (*color)[1], (*color)[2], (*color)[3]));
	  state_changed = true;
	}
      }
      
      
      if (node.GetProperty("maps")) {
	VectorObjectPointer* maps = node.GetProperty("maps")->AsVectorObjectPointer();
	if(maps) {
	  for(VectorObjectPointer::iterator i = maps->begin(); i != maps->end(); i++) {
	    MapObjectPointer* map = (*i)->AsMapObjectPointer();
	    File* name = map->GetProperty("name")->AsFile();

	    std::cout << "Map name " << map->GetProperty("name")->AsFile()->GetName() << std::endl;
	    if(name->GetName() == "diffuse" && node.GetNbPolygons() > 0) {

	      std::cout << "Map file " << map->GetProperty("file")->AsFile()->GetName() << std::endl;

	      std::string path = map->GetProperty("file")->AsFile()->GetName();
	      osg::Image* image = osgDB::readImageFile(path);



	      // not found so try in current directory
	      if (!image) {
		std::string::size_type p;
		std::string mapName;
		std::cerr << "Image file " << path << " not found" << std::endl;
		p=path.rfind("/");
		if (p!=std::string::npos) {
		  mapName=path.substr(p+1);
		  image = osgDB::readImageFile(path.substr(p+1));
		  if (!image)
		    std::cerr << "Image file " << mapName << " not found" << std::endl;
		} else {
		  p=path.rfind("\\");
		  mapName=path.substr(p+1);
		}

		image = osgDB::readImageFile(mapName);
		if (!image) {
		  mapName=mSourceData+mapName;
		  image = osgDB::readImageFile(mapName);
		  if (!image)
		    std::cerr << "Image file " << mapName << " not found" << std::endl;
		}
	      }


	      osg::Texture2D* texture = new osg::Texture2D;
	      texture->setImage(image);
	      state->setTextureAttributeAndModes(mTextures.size(), texture, osg::StateAttribute::ON);
	      mTextures.push_back(map);

	      VectorFloat* element = map->GetProperty("element")->AsVectorFloat();
	      int coord_size = int((*element)[0]);
	      osg::Array* array = 0;
	      switch(coord_size) {
	      case 2:
		array = new osg::Vec2Array;
		break;
	      case 3:
		array = new osg::Vec3Array;
		break;
	      default:
		std::cerr << "element with value " << coord_size << " unexpected" << std::endl;
		break;
	      }

	      mGeometry->setTexCoordArray(0, array);
	      state_changed = true;
	    }
	  }
	}
      }
      if(state_changed) mGeometry->setStateSet(state.get());

      if(node.GetNbPolygons() > 0) {
	mGeometry->setVertexArray(new osg::Vec3Array());

	if(node.find("normal") != node.end()) {
	  mGeometry->setNormalBinding(osg::Geometry::BIND_PER_VERTEX);
	  mGeometry->setNormalArray(new osg::Vec3Array);
	}
      }

      mGeode->addDrawable(mGeometry);

      Traverse(node);

      if(node.GetNbPolygons() > 0) {
	mGeometry->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::TRIANGLES, 0, mGeometry->getVertexArray()->getNumElements()));
      }

      mTextures.clear();
      mGeometry = 0;
    }

    void Apply(Vertex& node) {
      // visited in Apply(Polygon& node)
    }

    void Apply(Point& node) {
      // Explored by Apply(Polygon& node)
    }

    void Apply(Polygon& node) {
      if (!mGeometry)
	return;
      VectorObjectPointer* vertexes = node.GetProperty("vertexes")->AsVectorObjectPointer();
      for(VectorObjectPointer::iterator i = vertexes->begin(); i != vertexes->end(); i++) {
	Vertex* vertex = (*i)->AsVertex();

	for(unsigned int j = 0; j < mTextures.size(); j++) {
	  MapObjectPointer* map = mTextures[j];
	  osg::Array* texcoord = mGeometry->getTexCoordArray(j);
	  if(texcoord) {
	    std::string name = map->GetProperty("name")->AsFile()->GetName();
	    VectorFloat* coord = vertex->GetProperty(name)->AsVectorFloat();
	    switch(texcoord->getType()) {
	    case osg::Array::Vec2ArrayType:
	      {
		osg::Vec2Array* array = dynamic_cast<osg::Vec2Array*>(texcoord);
		array->push_back(osg::Vec2((*coord)[0], (*coord)[1]));
	      }
	    break;
	    case osg::Array::Vec3ArrayType:
	      {
		osg::Vec3Array* array = dynamic_cast<osg::Vec3Array*>(texcoord);
		array->push_back(osg::Vec3((*coord)[0], (*coord)[1], (*coord)[2]));
	      }
	    break;
	    default:
	      std::cerr << "array with type " << texcoord->getType() << " unexpected" << std::endl;
	      break;
	    }
	  }
	}

	osg::Vec3Array* normalarray = mGeometry->getNormalArray();
	if(normalarray) {
	  VectorFloat* normal = vertex->GetProperty("normal")->AsVectorFloat();
	  normalarray->push_back(osg::Vec3((*normal)[0], (*normal)[1], (*normal)[2]));
	}

	Vector3f position = vertex->GetPoint()->GetPosition();
	osg::Vec3Array* array = dynamic_cast<osg::Vec3Array*>(mGeometry->getVertexArray());
	osg::Vec3 vector = osg::Vec3(position[0], position[1], position[2]);
	array->push_back(vector);
	
      }

      Traverse(node);
    }

    void Apply(Mesh& node) {
      mGeode = new osg::Geode;

      Traverse(node);
    }

    void Apply(File& node) {
    }

    osg::Node* GetNode() { return mGeode; }

  private:
    std::vector<MapObjectPointer*> mTextures;
    osg::Geode* mGeode;
    osg::Geometry* mGeometry;
    std::vector<std::string> mArguments;
    std::string mSourceData;
  };


  osg::Node* PyEMSH2OSG(Object* object,std::vector<std::string> args) {
    EXGConverter converter;
    converter.SetArgs(args);
    object->Accept(converter);
    return converter.GetNode();
  }




  class OSGConverter : public osg::NodeVisitor {
  public:
    OSGConverter() : osg::NodeVisitor(osg::NodeVisitor::TRAVERSE_ALL_CHILDREN), mPyObject(0), mPyRoot(0) {
      mPyNodePath.clear();
    }

    void apply(osg::Node& node) {
      if(mPyObject) {
	if(mPyNodePath.empty()) {
	  mPyRoot = mPyObject;
	} else {
	  PyObject* parent = mPyNodePath.back();
	  if(PyObject_CallMethod(parent, "addNode", "O", mPyObject) == NULL)
	    abort();
	}
	mPyObject = 0;
      } else {
	std::cerr << "exg::OSGConverter:: " << node.className() << " not implemented" << std::endl;
      }
    }

    void apply(osg::Group& node) {
      if(!mPyObject)
	mPyObject = mScene.NodeNew("NodeGroup", node.getName());

      PyObject* instance = mPyObject;

      PyObject* children = PyList_New(0);
      PyObject_SetAttrString(instance, "children", children);

      apply((osg::Node&)node);

      mPyNodePath.push_back(instance);
      traverse(node);
      mPyNodePath.pop_back();
    }

    void apply(osg::MatrixTransform& node) {
      mPyObject = mScene.NodeNew("NodeTransform", node.getName());
      apply((osg::Group&)node);
    }

    void apply(osg::Geode& node) {
      mPyObject = mScene.NodeNew("NodeMesh", node.getName());
      apply((osg::Node&)node);
    }

    PyObject* GetPyRoot(void) { return mPyRoot; }

  private:
    typedef std::vector<PyObject*> PyNodePath;

    PyScene mScene;
    PyNodePath mPyNodePath;
    PyObject* mPyObject;
    PyObject* mPyRoot;
  };
  
  PyObject* OSG2EXG(osg::Node* node) {
    OSGConverter converter;
    node->accept(converter);
    return converter.GetPyRoot();
  }
};
